'''
API для работы с синхронизатором lsync 
v.0.2.0.1
-------------------------------------------------------------------------------
История изменений - в конце файла
OldBean, 14.08.2019
'''
import redis
#-------------------------------------------------------------------------------
# Сoeдиняемся с БД
rdb = redis.StrictRedis(host = 'localhost', port = 6379, db = 0)
# Список единиц измерения
units = [u.decode() for u in rdb.lrange('units', 0, -1)]
#-------------------------------------------------------------------------------
'''
Генерирование класса V, зависящего от конфигурации оборудования. Экземпляр
этого класса (v) позволяет получать или устанавливать значения физических
величин в БД через свойства класса.

Примеры использования:
v.T1 - значение температуры с 1-го датчика
v.q0 = 300 - установка скорости отбора через 0-й клапан на уровне 300 мл/час
v.Q0 - получить интеграл по времени от q0 (для клапана отбора - расход) с
    момента последнего сброса интегратора.
v.Q0 = 0 - сброс интегратора (расхода) клапана отбора 0
v.w0 = 600 - установка мощности нагрева ТЭНа, подключенного к 0-ому контроллеру
    PDM без учета сетевого напряжения (предполагается, что в сети 220В)
v.w0 = [600, 237] - то же самое, только с учетом напряжения сети. Здесь 237 - 
    RMS напряжения сети, измеренное датчиком RMS 
'''
codeV = '''\nclass V(object):'''
for key in [k.decode() for k in rdb.hkeys('calibr')]:
  codeV += '''\n  @property
  def ''' + key + '''(self):
    return float(rdb.lindex(\'''' + key + '''\', -1))
    print('Устройства с ключем %s нет в БД'%(\'''' + key + '''\'))
'''
  if key[0] in ('q', 'r', 'w'): # Исполнительные устройства
    codeV += '''\n  @''' + key + '''.setter''' + '''
  def ''' + key + '''(self, value):
    if type(value) in (tuple, list):
      coef = 220.0/value[1]
      rdb.lset(\'''' + key + '''\', -1, value[0]*coef*coef)
    else:
      rdb.lset(\'''' + key + '''\', -1, value)
    rdb.rpush(\'must\', \'''' + key + '''\')
'''
  elif key[0] == 'Q': # Сброс интеграторов PWM-контроллеров 
    codeV += '''\n  @''' + key + '''.setter''' + '''
  def ''' + key + '''(self, value):
    rdb.rpush('reset', \'''' + key + '''\')
'''
ccodeV = compile(codeV, '', 'exec') 
exec(ccodeV)
v = V()
#-------------------------------------------------------------------------------
'''
Генерирование класса С, зависящего от конфигурации оборудования. Экземпляр
этого класса (с) позволяет получать или устанавливать калибровочные данные
датчиков и контроллеров, соответствующих физическим велинам в БД через
свойства класса. 

Сеттер должен получать калибровочные данные в виде списка [coef, offset, ui],
где вещественный параметр coef задает масштабирующий коэффициент в единицах
измерения, определяемой параметром ui (целочисленный индекс единицы измерения
в списке units), offset - целочисленное смещение кода АЦП. Конечная физическая
величина равна v = coef*(код АЦП - offset).

Геттер возвращается результат в виде списка [coef, offset, ui].
  
Примеры использования:
c.U0 возвращает калибровочные данные 0-го датчика RMS
c.q1 = [1015.0, 0, 3] - установить калибровочные данные 1-го клапана отбора.
    В данном случае 1015.0 - это скорость отбора полностью открытого клапана.
c.P0 = [0.0734, 38, 5] - установить калибровочные данные датчика давления
    MPX5010DP (у него ключ 'P0')
'''
codeC = '''\nclass C(object):'''
for key in [k.decode() for k in rdb.hkeys('calibr')]:
  if key in ('time'):
    continue
  codeC += '''\n  @property
  def ''' + key + '''(self):
    return deser(rdb.hget(\'calibr\', \'''' + key + '''\'))
'''
  codeC += '''\n  @''' + key + '''.setter''' + '''
  def ''' + key + '''(self, value):
    rdb.hset('calibr', \'''' + key + '''\', ser(*value))
    rdb.rpush(\'cmust\', \'''' + key + '''\')
'''
ccodeC = compile(codeC, '', 'exec') 
exec(ccodeC)
c = C()
#-------------------------------------------------------------------------------
'''
Через свойства класса U (utility) будет реализовываться дополнительный
функционал. На данный момент готовы: получение всех ключей физических величин,
обнуление (выключение) всех контроллеров и завершение работы синхронизатора.

Примеры использования:

u.keys - возвращает все ключи (доступные физические величины, соответствующие
    всем датчикам и контроллером + время)
u.off - выключение всех контроллеров (реле, PDM и PWM)
u.quit - завершение работы синхронизатора

'''
codeU = '''\nclass U(object):'''
codeU += '''\n  @property
  def keys(self):
    return [key.decode() for key in rdb.hkeys('calibr')]
'''
codeU += '''\n  @property
  def off(self):
    for key in [k.decode() for k in rdb.hkeys('calibr')]:
      if key[0] in ('q', 'r', 'w'): # Исполнительные устройства
        rdb.lset(key, -1, 0)
        rdb.rpush(\'must\', key)
'''
codeU += '''\n  @property
  def quit(self):
    self.off
    rdb.set(b\'status\', \'q\')
'''
ccodeU = compile(codeU, '', 'exec') 
exec(ccodeU)
u = U()
#-------------------------------------------------------------------------------
def ser(coef, offset, ui):
  '''
  ser(coef, offset, ui)
  Вспомогательная функция для упаковки (сериализации) калибровочных данных (в
  6 байтов) для хранения в БД. ТОчно в таком же виде эти данные хранятся в
  EEPROM устройств
  '''
  ip = int(coef) & 0xffff # Целая часть коэффициента
  fp = int((coef - ip)*10000) & 0xffff # Нормированная дробная часть
  buf = b''
  buf += ip.to_bytes(2, 'little')
  buf += fp.to_bytes(2, 'little')
  buf += (offset & 0xff).to_bytes(1, 'little')
  buf += (ui & 0xff).to_bytes(1, 'little')
  return buf
#-------------------------------------------------------------------------------
def deser(buf):
  '''
  deser(buf)
  Вспомогательная функция для распаковки калибровочных данных. Результат
  возвращается в виде списка [coef, offset, ui], где coef - масштабирующий
  коэффициент, offset - смещение и ui - индекс единицы измерения
  '''
  ip = int(buf[0]) # Младший байт целой части
  ip |= (int(buf[1])) << 8 # Старший байт целой части
  fp = int(buf[2]) # Младший байт дробной части
  fp |= (int(buf[3])) << 8 # Старший байт дробной части
  return [ip + fp*0.0001, int(buf[4]), int(buf[5])]
#===============================================================================
def uget(key):
  '''
  uget(key)
  Получить единицу измерения физической величины с ключем key
  Функция возвращет строку с единицей измерения, ассоциированой с физической
  величиной, хранимой в БД с ключем key. Все единицы измерения хранятся в
  списке БД под ключем 'units' (в данном модуле доступен список units). В
  калибровочных данных хранится индекс единицы измерения в этих списках.
  
  Примеры использования:
  uget('U0') - единица измерения напряжения сети, получаемого с 0-го датчика
    RMS ('В')
  uget('q2') - единица измерения скорости отбора 2-го клапана отбора (мл/час)
  
  '''
  if key.encode() in rdb.hkeys('calibr'):
    return units[deser(rdb.hget('calibr', key))[2]]
  else:
    print('Устройства с ключем %s нет в БД'%(key))
#-------------------------------------------------------------------------------
def cget(key):
  '''
  cget(key)
  Получение калибровочных данных устройства с ключем key.
  Результат возвращается в виде списка [coef, offset, ui], 
  где coef - масштабирующий коэффициент, offset - смещение и ui - индекс
  единицы измерения
  '''
  if key.encode() in rdb.hkeys('calibr'):
    return deser(rdb.hget('calibr', key))
  else:
    print('Устройства с ключем %s нет в БД'%(key))
#-------------------------------------------------------------------------------
def cset(key, coef, offset, ui):
  '''
  cset(key, coef, offset, ui)
  Запись калибровочных данных в устройство и БД. Вещетсвенный параметр coef 
  задает масштабирующий коэффициент в единицах измерения, определяемой
  параметром ui (целочисленный индекс единицы измерения в списке units),
  offset - целочисленное смещение кода АЦП. Конечная физическая величина
  равна v = coef*(код АЦП - offset)
  
  Примеры использования:
  cset('U0', 0.3839, 0, 1) - установить калибровочные данные 0-го датчика RMS
  cset('q1', 1015.0, 0, 3) - установить калибровочные данные 1-го клапана
    отбора. В данном случае 1015.0 - скорость отбора полностью открытого
    клапана
  cset('P0', 0.0734, 38, 5) - установить калибровочные данные датчика давления
    MPX5010DP с ключем 'P0'
  '''
  if key.encode() in rdb.hkeys('calibr'):
    rdb.hset('calibr', key, ser(coef, offset, ui))
    rdb.rpush('cmust', key) # Сообщение синхронизатору "Обновить калибровку!"
  else:
    print('Устройства с ключем %s нет в БД'%(key))
#-------------------------------------------------------------------------------
def temp(fon = 'temp.txt', start = -1800, end = -1):
  '''
  temp(fon = 'temp.txt', start = -1800, end = -1)
  Выгрузка в файл зависимостей всех температур от времени в виде таблицы.
  Параметр fon задает имя файла (по умолчанию 'temp.txt'), параметры start и 
  end задают временной интервал (по умолчанию - за истекший час)

  Значения температур и времени берутся из соответствующих списков БД redis.
  Параметры start и end должны быть отрицательными (start <= end). Они задают
  начальное и конечное значение индексов выбираемых из списков значений.
  В данной версии синхронизатора (0.2.0.0) данные записываются с интервалом
  2 сек. -1 означает текущий момент.
  
  Примеры использования:
  temp() - выгрузить все температуры за истекший час в файл 'temp.txt'
  temp('my_temps.txt', -900) - выгрузить все температуры за последние пол-часа
    в файл 'my_temps.txt'
  temp('temperatures', -3600, -1800) - за предыдущий час
  '''
  keys = []
  for key in rdb.hkeys('calibr'):
    if key[0] == ord('T') or key == b'time':
      keys.append(key)
  data = []
  for key in keys:
    data.append(rdb.lrange(key, start, end))
  with open(fon, 'w') as fo:
    s = ''
    for key in keys:
      s += key.decode() + ' '
    s += '\n'
    fo.write(s)
    for i in range(len(data[0])):
      s = ''
      for j in range(len(data)):
        s += data[j][i].decode() + ' '
      s += '\n'
      fo.write(s)
#-------------------------------------------------------------------------------
def keys():
  '''
  keys()
  Получение списка ключей всех устройств (контроллеров и датчиков)
  '''
  return [key.decode() for key in rdb.hkeys('calibr')]
#-------------------------------------------------------------------------------
def vget(key, start = -1, end = -1):
  '''
  vget(key, start = -1, end = -1)
  Получение значения или списка значений датчика с ключем key.
  Значения берутся из соответствующего списка БД redis. Параметры start и end
  (отрицательные, start <= end) задают начальный и конечный индексы списка.
  Если start = end возвращается одно значение. В противном случае возвращается
  список значений в заданном интервале. -1 означает текущее (последнее)
  значение, сораненное в БД.
  
  Примеры использования:
    vget('P0') - текущее значение 0-го датчика давления
    vget('T2', -100) - последние 100 значений 2-го датчика температуры, 
        с шагом по времени - dt. В данной версии синхронизатора dt = 2 сек
    vget('T0', -1100, -1090) - 10 значений 0-го датчика температуры, начиная
        с момента времени 1100*dt до текущего момента (приблизительно).
    vget('Q2') - суммарный расход через 2-клапан отбора с момента последнего
      сброса (см.функцию resetQ)
  '''
  if start > end:
    print('Начальный момент должен быть раньше конечного')
    return []
  if key.encode() in rdb.hkeys('calibr'):
    if start == end:
      return float(rdb.lindex(key, start))
    else:
      return [float(v) for v in rdb.lrange(key, start, end)]
  else:
    print('Устройства с ключем %s нет в БД'%(key))
#-------------------------------------------------------------------------------
def vset(key, value, rms = None):
  '''
  vset(key, value, rms = None)
  Установка значение устройства (контроллера) с ключем key.
  Параметр value задает значение в единицах измерения, которые определены в
  калибровочных данных устройства (могут быть получены вызовом uget(key)).
  Для контроллеров PDM предусмотрена коррекция устанавливаемой мощности по
  RMS напряжения сети (стабилизация мощности нагрева). Для этого нужно задать
  значение параметра rms, равное текущему RMS напряжения сети (в вольтах),
  измеренное соответствующим датчиком RMS.
  
  Примеры использования:
  
    vset('r1', 1) - включить 1-е реле
    vset('r2', 0) - выключить 2-е реле
    vset('q0', 350) - установить скорость отбора 0-го клапана на 350 мл/час
    vset('w0', 600) - установить мощность нагревателя, подключенного к 0-му
      контроллеру PDM на 600 Вт (по номиналу, без учета напряжения в сети)
    vset('w0', 600, 232) - установить мощность нагревателя, подключенного к 
      0-му контроллеру PDM на 600 Вт с учетом напряжения сети. Здесь 232 - 
      RMS напряжения сети, измеренное датчиком RSM.
    vset('q2', 0) - выключить 2-й клапан отбора
  '''
  if key.encode() in rdb.hkeys('calibr'):
    if not (rms is None): # Нужна коррекция value на RMS напряжения сети
      coef = 220.0/rms
      rdb.lset(key, -1, value*coef*coef)
    else:
      rdb.lset(key, -1, value) # Изменяем последнее значение в списке БД
    rdb.rpush('must', key) # Сообщение синхронизатору "Обновить состояние!"
  else:
    print('Устройства с ключем %s нет в БД'%(key))
#-------------------------------------------------------------------------------
def resetQ(key): # Сброс интегратора контроллера PWM с ключем key
  '''
  resetQ(key)
  Сброс интеграторов, подсчитывающих интеграл по времени величины, регулируемой
  контроллером PWM с ключем key. В случае, если к контроллеру подкючен клапан
  отбора - это суммарный расход через данный клапан. 
  '''
  if key.encode() in rdb.hkeys('calibr'):
    rdb.rpush('reset', key)
  else:
    print('Устройства с ключем %s нет в БД'%(key))
#-------------------------------------------------------------------------------
def quit():
  '''
  quit()
  Завершение работы синхронизатора
  '''
  rdb.set(b'status', 'q')
#-------------------------------------------------------------------------------
'''
12.08.2019 1. Реализованы и отлажены функции: keys, vget, vset, uget, resetQ,
                                              cget, cset, temp и quit
           2. Сделано их описание

14.08.2019 1. Реализованы динамические классы V, C и U, позволяющие работать
              с БД через свойства классов.
'''
